import base64
import gzip
import pickle
import tempfile
from contextlib import contextmanager
from datetime import datetime
from functools import cached_property
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Literal, Optional, Tuple, Union

import magic
from pikepdf import Pdf
from PIL import Image, ImageDraw
from pydantic import BaseModel, Field, PositiveInt, PrivateAttr, computed_field, field_serializer, field_validator

from docprompt._exec.ghostscript import compress_pdf_to_bytes, rasterize_page_to_bytes, rasterize_pdf_to_bytes
from docprompt.schema.operations import PageTextExtractionOutput

from .layout import TextBlock

if TYPE_CHECKING:
    from docprompt.service_providers.base import BaseProvider

DEFAULT_DPI = 100


def get_page_render_size_from_bytes(file_bytes: bytes, page_number: int, dpi: int = DEFAULT_DPI):
    """
    Returns the render size of a page in pixels
    """
    import pdfplumber

    with pdfplumber.PDF.open(BytesIO(file_bytes)) as pdf:
        page = pdf.pages[page_number]

        width_pt, height_pt = page.mediabox.upper_right

    width_in = width_pt / 72
    height_in = height_pt / 72

    width_px = int(width_in * dpi)
    height_px = int(height_in * dpi)

    return (width_px, height_px)


class Document(BaseModel):
    """
    Represents a PDF document
    """

    name: str = Field(description="The name of the document")
    file_bytes: bytes = Field(description="The bytes of the document", repr=False)
    file_path: Optional[str] = None
    text_sidecars: Dict[str, Dict[int, PageTextExtractionOutput]] = Field(default_factory=dict, repr=False)

    _raster_cache: Dict[int, Dict[int, bytes]] = PrivateAttr(default_factory=dict)

    def __len__(self):
        return len(self.pages)

    def __hash__(self):
        return hash(self.document_hash)

    @computed_field
    @cached_property
    def page_count(self) -> PositiveInt:
        from docprompt.utils.util import get_page_count

        return get_page_count(self.file_bytes)

    @property
    def num_pages(self):
        return self.page_count

    @computed_field
    @cached_property
    def document_hash(self) -> str:
        from docprompt.utils.util import hash_from_bytes

        return hash_from_bytes(self.file_bytes)

    @field_serializer("file_bytes")
    def serialize_file_bytes(self, v: bytes, _info):
        compressed = gzip.compress(v)

        return base64.b64encode(compressed).decode("utf-8")

    @field_validator("file_bytes")
    def validate_file_bytes(cls, v: bytes):
        if not isinstance(v, bytes):
            raise ValueError("File bytes must be bytes")

        if len(v) == 0:
            raise ValueError("File bytes must not be empty")

        if magic.from_buffer(v, mime=True) == "text/plain":
            v = base64.b64decode(v, validate=True)

        if magic.from_buffer(v, mime=True) == "application/gzip":
            v = gzip.decompress(v)

        if magic.from_buffer(v, mime=True) != "application/pdf":
            raise ValueError("File bytes must be a PDF")

        return v

    @classmethod
    def from_path(cls, file_path: Union[PathLike, str]):
        file_path = Path(file_path)

        if not file_path.is_file():
            raise ValueError(f"File path {file_path} is not a file")

        file_bytes = file_path.read_bytes()

        return cls(name=file_path.name, file_path=str(file_path), file_bytes=file_bytes)

    @classmethod
    def from_bytes(cls, file_bytes: bytes, name: Optional[str] = None):
        if name is None:
            name = f"PDF-{datetime.now().isoformat()}.pdf"

        return cls(name=name, file_bytes=file_bytes)

    def get_bytes(self) -> bytes:
        return self.file_bytes  # Deprecated

    @property
    def path(self):
        return self.file_path

    def get_page_render_size(self, page_number: int, dpi: int = DEFAULT_DPI) -> Tuple[int, int]:
        """
        Returns the render size of a page in pixels
        """
        return get_page_render_size_from_bytes(self.get_bytes(), page_number, dpi=dpi)

    def to_compressed_bytes(self, compression_kwargs: dict = {}) -> bytes:
        """
        Compresses the document using Ghostscript
        """
        with self.as_tempfile() as temp_path:
            return compress_pdf_to_bytes(temp_path, **compression_kwargs)

    def rasterize_page(
        self,
        page_number: int,
        *,
        dpi: int = DEFAULT_DPI,
        downscale_size: Optional[Tuple[int, int]] = None,
        device="png16m",
        use_cache: bool = True,
    ) -> bytes:
        """
        Rasterizes a page of the document using Ghostscript
        """
        generated_image = False

        if page_number < 0 or page_number > self.num_pages:
            raise ValueError(f"Page number must be between 0 and {self.num_pages}")

        if use_cache and self._raster_cache.get(dpi, {}).get(page_number):
            rastered = self._raster_cache[dpi][page_number]
        else:
            with self.as_tempfile() as temp_path:
                rastered = rasterize_page_to_bytes(temp_path, page_number, dpi=dpi, device=device)
                generated_image = True

        if use_cache and generated_image:
            self._raster_cache.setdefault(dpi, {})
            self._raster_cache[dpi][page_number] = rastered

        if downscale_size:
            with Image.open(BytesIO(rastered)) as img:
                width, height = img.size
                if not (width < downscale_size[0] or height < downscale_size[1]):
                    img = img.resize(downscale_size)
                    img_bytes = BytesIO()
                    img.save(img_bytes, format="PNG")
                    rastered = img_bytes.getvalue()

        return rastered

    def rasterize_page_to_data_uri(
        self,
        page_number: int,
        *,
        dpi: int = DEFAULT_DPI,
        downscale_size: Optional[Tuple[int, int]] = None,
        device="png16m",
        use_cache: bool = True,
    ) -> str:
        """
        Rasterizes a page of the document using Ghostscript and returns a data URI, which can
        be embedded into HTML or passed to large language models
        """
        image_bytes = self.rasterize_page(
            page_number, dpi=dpi, downscale_size=downscale_size, device=device, use_cache=use_cache
        )
        return f"data:image/png;base64,{base64.b64encode(image_bytes).decode('utf-8')}"

    def rasterize_pdf(
        self, dpi: int = DEFAULT_DPI, device="pnggray", downscale_factor: Optional[int] = None, use_cache: bool = True
    ) -> Dict[int, bytes]:
        """
        Rasterizes the entire document using Ghostscript
        """
        if use_cache and self._raster_cache.get(dpi) and len(self._raster_cache[dpi]) == self.num_pages:
            return self._raster_cache[dpi]

        with self.as_tempfile() as temp_path:
            result = rasterize_pdf_to_bytes(temp_path, dpi=dpi, device=device, downscale_factor=downscale_factor)

        if use_cache:
            self._raster_cache[dpi] = result.copy()  # Shallow copy should be OK

        return result

    def split(self, start: Optional[int] = None, stop: Optional[int] = None):
        """
        Splits a document into multiple documents
        """
        if start is None and stop is None:
            raise ValueError("Must specify either start or stop")

        start = start or 0

        from docprompt.utils.splitter import split_pdf_to_bytes

        split_bytes = split_pdf_to_bytes(self.file_bytes, start_page=start, stop_page=stop)

        return Document.from_bytes(split_bytes, name=self.name)

    def as_tempfile(self, **kwargs):
        """
        Returns a tempfile of the document
        """

        @contextmanager
        def tempfile_context() -> str:
            tempfile_kwargs = {"mode": "wb", "delete": True, "suffix": ".pdf", **kwargs}

            with tempfile.NamedTemporaryFile(**tempfile_kwargs) as f:
                f.write(self.file_bytes)
                f.flush()
                yield f.name

        return tempfile_context()

    def write_to_path(self, path: Union[PathLike, str], **kwargs):
        """
        Writes the document to a path
        """
        path = Path(path)

        if path.is_dir():
            path = path / self.name

        with path.open("wb") as f:
            f.write(self.file_bytes)

    @property
    def text_data(self):
        try:
            return next(iter(self.text_sidecars.values()))
        except StopIteration:
            return None

    def perform_text_extraction(
        self, provider: "BaseProvider", cache: bool = True
    ) -> Dict[int, PageTextExtractionOutput]:
        """
        Performs text extraction for a given provider
        """

        result = provider.process_document(self)

        sidecars = {}

        if not result or not result.page_results:
            return sidecars

        for page_result in result.page_results:
            if not page_result.ocr_result:
                continue

            sidecars[page_result.page_number] = page_result.ocr_result

        if cache:
            self.text_sidecars[provider.name] = sidecars

        return sidecars



import multiprocessing
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from threading import Lock
from typing import TYPE_CHECKING, Dict, Literal, Optional, Union

import tqdm
from tenacity import retry, stop_after_attempt, wait_exponential

from docprompt.schema.document import Document
from docprompt.schema.layout import BoundingPoly, NormBBox, Point, SegmentLevels, TextBlock, TextSpan
from docprompt.schema.operations import PageResult, PageTextExtractionOutput
from docprompt.service_providers.base import ProviderResult
from docprompt.service_providers.types import OPERATIONS
from docprompt.utils.splitter import pdf_split_iter_with_max_bytes

from .base import BaseProvider, ProviderResult

if TYPE_CHECKING:
    from google.cloud import documentai

    from docprompt.schema.document import Document


orientation_rotation_mapping = {
    0: 0,
    1: 0,
    2: 90,
    3: 180,
    4: -90,
}

service_account_file_read_lock = Lock()

# This will wait up to ~8 minutes before giving up, which covers almost all high-contention cases
# TODO: Scope this to only retry on 429 and 5xx
default_retry_decorator = retry(wait=wait_exponential(multiplier=1, max=60), stop=stop_after_attempt(10))


def bounding_poly_from_layout(layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"]):
    return BoundingPoly(
        normalized_vertices=[
            Point(x=round(vertex.x, 5), y=round(vertex.y, 5)) for vertex in layout.bounding_poly.normalized_vertices
        ]
    )


def bounding_box_from_layout(layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"]):
    sorted_vertices = sorted(layout.bounding_poly.normalized_vertices, key=lambda x: (x.x, x.y))
    upper_left = sorted_vertices[0]
    lower_right = sorted_vertices[-1]

    return NormBBox(
        x0=round(upper_left.x, 5),
        top=round(upper_left.y, 5),
        x1=round(lower_right.x, 5),
        bottom=round(lower_right.y, 5),
    )


def geometry_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
    exclude_bounding_poly: bool = False,
):
    bounding_poly = None if exclude_bounding_poly else bounding_poly_from_layout(layout)

    bounding_box = bounding_box_from_layout(layout)

    return {
        "bounding_poly": bounding_poly,
        "bounding_box": bounding_box,
    }


def text_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
    document_text: str,
    offset: int = 0,
) -> str:
    """
    Offset is used to account for the fact that text references
    are relative to the entire document.
    """
    working_text = ""

    for segment in sorted(layout.text_anchor.text_segments, key=lambda x: x.end_index):
        start = getattr(segment, "start_index", 0)
        end = segment.end_index

        working_text += document_text[start - offset : end - offset]

    return working_text


def text_spans_from_layout(
    layout: Union["documentai.Document.Page.Layout", "documentai.Document.Page.Token"],
    level: Literal["page", "document"],
    offset: int = 0,
) -> list[TextSpan]:
    text_spans = []

    for segment in sorted(layout.text_anchor.text_segments, key=lambda x: x.end_index):
        start = getattr(segment, "start_index", 0)
        end = segment.end_index

        text_spans.append(
            TextSpan(
                start_index=start - offset,
                end_index=end - offset,
                level=level,
            )
        )

    return text_spans


def text_blocks_from_page(
    page: "documentai.Document.Page",
    document_text: str,
    type: Literal["line", "block", "token", "paragraph"],
    *,
    exclude_bounding_poly: bool = False,
) -> list[TextBlock]:
    text_blocks = []

    type_mapping: Dict[str, SegmentLevels] = {
        "line": "line",
        "paragraph": "paragraph",
        "block": "block",
        "token": "word",
    }

    orientation_mapping = {
        1: "UP",
        2: "RIGHT",
        3: "DOWN",
        4: "LEFT",
    }

    # Offset is used to account for the fact that text references are relative to the entire document.
    # while we need to compute spans relative to the page.
    offset_low = page.layout.text_anchor.text_segments[0].start_index or 0

    for item in getattr(page, f"{type}s"):
        layout = item.layout
        block_text = text_from_layout(layout, document_text)
        geometry_kwargs = geometry_from_layout(layout, exclude_bounding_poly=exclude_bounding_poly)
        confidence = layout.confidence
        orientation = orientation_mapping.get(layout.orientation, "UP")

        text_spans = text_spans_from_layout(layout, level="page", offset=offset_low)

        block_type = type_mapping[type]
        text_blocks.append(
            TextBlock(
                text=block_text,
                type=block_type,
                bounding_box=geometry_kwargs["bounding_box"],
                bounding_poly=geometry_kwargs["bounding_poly"],
                confidence=round(confidence, 5),
                direction=orientation,
                text_spans=text_spans,
            )
        )

    return text_blocks


def process_page(
    document_text: str, page, doc_page_num: int, provider_name: str, exclude_bounding_poly: bool = False
) -> PageResult:
    layout = page.layout

    page_text = text_from_layout(layout, document_text)

    word_boxes = text_blocks_from_page(page, document_text, "token", exclude_bounding_poly=exclude_bounding_poly)
    line_boxes = text_blocks_from_page(page, document_text, "line", exclude_bounding_poly=exclude_bounding_poly)
    block_boxes = text_blocks_from_page(page, document_text, "block", exclude_bounding_poly=exclude_bounding_poly)

    ocr_result = PageTextExtractionOutput(
        text=page_text,
        words=word_boxes,
        lines=line_boxes,
        blocks=block_boxes,
    )

    return PageResult(
        provider_name=provider_name,
        page_number=doc_page_num,
        ocr_result=ocr_result,
    )


def gcp_documents_to_result_single(
    documents: list["documentai.Document"], provider_name: str, *, exclude_bounding_poly: bool = False
):
    page_offset = 1  # We want pages to be 1-indexed

    page_results = []

    for document in tqdm.tqdm(documents):
        for doc_page_num, page in enumerate(document.pages):
            page_results.append(
                process_page(
                    document.text,
                    page,
                    page_offset + doc_page_num,
                    provider_name,
                    exclude_bounding_poly=exclude_bounding_poly,
                )
            )

        page_offset += doc_page_num + 1

    return ProviderResult(
        provider_name="GoogleDocumentAIProvider",
        page_results=page_results,
    )


def multi_process_page(args):
    idx, document, provider_name, page_offset, exclude_bounding_poly = args

    page_results = []

    for doc_page_num, page in enumerate(document.pages):
        page_results.append(
            process_page(
                document.text,
                page,
                page_offset + doc_page_num,
                provider_name,
                exclude_bounding_poly=exclude_bounding_poly,
            )
        )

    return idx, page_results


def gcp_documents_to_result_multi(
    documents: list["documentai.Document"], provider_name: str, *, exclude_bounding_poly: bool = False
):
    tasks = []
    page_offset = 1  # Pages are 1-indexed

    # Prepare tasks for processing, each with a document index
    for idx, document in enumerate(documents):
        args = (
            idx,
            document,
            provider_name,
            page_offset,
            exclude_bounding_poly,
        )
        tasks.append(args)
        page_offset += len(document.pages)

    document_results = [None] * len(documents)

    worker_count = min(len(documents), min(4, max(multiprocessing.cpu_count(), 1)))

    futures = []

    with tqdm.tqdm(total=len(documents), desc="Processing documents") as pbar:
        with ProcessPoolExecutor(max_workers=worker_count) as executor:
            for task in tasks:
                future = executor.submit(multi_process_page, task)
                future.add_done_callback(lambda x: pbar.update(1))
                futures.append(future)

            for future in as_completed(futures):
                idx, page_results = future.result()
                document_results[idx] = page_results

    return ProviderResult(
        provider_name="GoogleDocumentAIProvider",
        page_results=[page for doc_pages in document_results for page in doc_pages],
    )


def gcp_documents_to_result(
    documents: list["documentai.Document"], provider_name: str, *, exclude_bounding_poly: bool = False
) -> ProviderResult:
    if True or len(documents) == 1 or multiprocessing.cpu_count() == 1:
        print("Using single process")
        return gcp_documents_to_result_single(
            documents,
            provider_name,
            exclude_bounding_poly=exclude_bounding_poly,
        )
    else:
        print("Using multiprocessing")
        return gcp_documents_to_result_multi(
            documents,
            provider_name,
            exclude_bounding_poly=exclude_bounding_poly,
        )


class GoogleDocumentAIProvider(BaseProvider):
    name = "GoogleDocumentAIProvider"

    max_bytes_per_request = 1024 * 1024 * 20  # 20MB is the max size for a single sync request
    max_page_count = 15

    def __init__(
        self,
        project_id: str,
        processor_id: str,
        *,
        service_account_info: Optional[dict] = None,
        service_account_file: Optional[str] = None,
        location: str = "us",
        max_workers: int = multiprocessing.cpu_count() * 2,
        exclude_bounding_poly: bool = False,
    ):
        if service_account_info is None and service_account_file is None:
            raise ValueError("You must provide either service_account_info or service_account_file")
        if service_account_info is not None and service_account_file is not None:
            raise ValueError("You must provide either service_account_info or service_account_file, not both")

        self.project_id = project_id
        self.processor_id = processor_id
        self.location = location

        self.max_workers = max_workers

        self.service_account_info = service_account_info
        self.service_account_file = service_account_file

        self.exclude_bounding_poly = exclude_bounding_poly

        try:
            from google.cloud import documentai

            self.documentai = documentai
        except ImportError:
            raise ImportError(
                "Please install 'google-cloud-documentai' to use the GoogleCloudVisionTextExtractionProvider"
            )

    def get_documentai_client(self, client_option_kwargs: dict = {}, **kwargs):
        from google.api_core.client_options import ClientOptions

        opts = ClientOptions(
            **{
                "api_endpoint": "us-documentai.googleapis.com",
                **client_option_kwargs,
            }
        )

        base_service_client_kwargs = {
            **kwargs,
            "client_options": opts,
        }

        if self.service_account_info is not None:
            return self.documentai.DocumentProcessorServiceClient.from_service_account_info(
                info=self.service_account_info,
                **base_service_client_kwargs,
            )
        elif self.service_account_file is not None:
            with service_account_file_read_lock:
                return self.documentai.DocumentProcessorServiceClient.from_service_account_file(
                    filename=self.service_account_file,
                    **base_service_client_kwargs,
                )
        else:
            raise ValueError("Missing account info and service file path.")

    @property
    def capabilities(self) -> list[OPERATIONS]:
        return [
            OPERATIONS.TEXT_EXTRACTION,
            OPERATIONS.LAYOUT_ANALYSIS,
            OPERATIONS.IMAGE_PROCESSING,
        ]

    def _process_document_sync(self, file_bytes: bytes):
        """
        Split the document into chunks of 15 pages or less, and process each chunk
        synchronously.
        """
        client = self.get_documentai_client()
        processor_name = client.processor_path(
            project=self.project_id,
            location=self.location,
            processor=self.processor_id,
        )

        documents = []

        @default_retry_decorator
        def process_byte_chunk(split_bytes: bytes):
            raw_document = self.documentai.RawDocument(
                content=split_bytes,
                mime_type="application/pdf",
            )

            field_mask = "text,pages.layout,pages.words,pages.lines,pages.tokens,pages.blocks"

            request = self.documentai.ProcessRequest(
                name=processor_name, raw_document=raw_document, field_mask=field_mask
            )

            result = client.process_document(request=request)

            return result.document

        with tqdm.tqdm(total=len(file_bytes), unit="B", unit_scale=True, desc="Processing document") as pbar:
            for split_bytes in pdf_split_iter_with_max_bytes(
                file_bytes, max_page_count=self.max_page_count, max_bytes=self.max_bytes_per_request
            ):
                document = process_byte_chunk(split_bytes)

                documents.append(document)

                pbar.update(len(split_bytes))

        return gcp_documents_to_result(documents, self.name, exclude_bounding_poly=self.exclude_bounding_poly)

    def _process_document_concurrent(self, file_bytes: bytes):
        # Process page chunks concurrently
        client = self.get_documentai_client()
        processor_name = client.processor_path(
            project=self.project_id,
            location=self.location,
            processor=self.processor_id,
        )

        print("Splitting document into chunks...")
        document_byte_splits = list(
            pdf_split_iter_with_max_bytes(
                file_bytes, max_page_count=self.max_page_count, max_bytes=self.max_bytes_per_request
            )
        )

        max_workers = min(len(document_byte_splits), self.max_workers)

        @default_retry_decorator
        def process_byte_chunk(split_bytes: bytes):
            raw_document = self.documentai.RawDocument(
                content=split_bytes,
                mime_type="application/pdf",
            )

            field_mask = "text,pages.layout,pages.words,pages.lines,pages.tokens,pages.blocks"

            request = self.documentai.ProcessRequest(
                name=processor_name, raw_document=raw_document, field_mask=field_mask
            )

            result = client.process_document(request=request)

            document = result.document

            return document

        print(f"Processing {len(document_byte_splits)} chunks...")
        with tqdm.tqdm(total=len(document_byte_splits), desc="Processing document") as pbar:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                future_to_index = {
                    executor.submit(process_byte_chunk, split): index
                    for index, split in enumerate(document_byte_splits)
                }

                documents = [None] * len(document_byte_splits)

                for future in as_completed(future_to_index):
                    index = future_to_index[future]
                    documents[index] = future.result()
                    pbar.update(1)

        print("Recombining")
        return gcp_documents_to_result(documents, self.name, exclude_bounding_poly=self.exclude_bounding_poly)

    def _call(self, document: Document, pages=...) -> ProviderResult:
        return self._process_document_concurrent(document.get_bytes())



import multiprocessing
import tempfile
from os import PathLike
from pathlib import Path
from subprocess import PIPE, CompletedProcess, run
from typing import Dict, List, Literal, Optional, Union

GS = "gs"

_RENDER_THREAD_COUNT = max(min(multiprocessing.cpu_count() - 2, 8), 1)


class GhostscriptError(Exception):
    def __init__(self, message: str, process: CompletedProcess) -> None:
        self.process = process
        super().__init__(message)


def _validate_device(device: str) -> str:
    if device not in ["pnggray", "png16m", "png256"]:
        raise ValueError("Invalid device")

    return device


def rasterize_page(
    fp: Union[PathLike, str],
    output_path: str,
    idx: int,
    *,
    dpi: int = 200,
    device="pnggray",
):
    device = _validate_device(device)
    args = [
        GS,
        "-q",
        "-dSAFER",
        "-dBATCH",
        "-dNOPAUSE",
        "-sDEVICE=png16m",
        "-dTextAlphaBits=4",
        "-dGraphicsAlphaBits=4",
        f"-sDEVICE={device}",
        f"-dFirstPage={idx}",
        f"-dLastPage={idx}",
        f"-r{dpi}",
        f"-sOutputFile={output_path}",
        "-f",
        str(fp),
    ]

    result = run(args, stdout=PIPE, stderr=PIPE, check=False)

    if result.returncode != 0:
        raise GhostscriptError("Ghostscript failed to rasterize the document: ", result)

    return result


def rasterize_pdf(
    fp: Union[PathLike, str],
    output_path: str,
    *,
    dpi: int = 100,
    device="pnggray",
    downscale_factor: Optional[int] = None,
):
    device = _validate_device(device)
    base_args = [
        GS,
        "-q",
        "-dSAFER",
        "-dBATCH",
        "-dNOPAUSE",
        "-dTextAlphaBits=4",
        "-dGraphicsAlphaBits=4",
        "-dBufferSpace=250000000",  # 250 Mb of buffer space.
        f"-dNumRenderingThreads={_RENDER_THREAD_COUNT}",
    ]

    if downscale_factor is not None:
        base_args += [f"-dDownScaleFactor={downscale_factor}"]

    args = base_args + [
        f"-sDEVICE={device}",
        f"-r{dpi}",
        f"-sOutputFile={output_path}",
        "-f",
        str(fp),
    ]

    result = run(args, stdout=PIPE, stderr=PIPE, check=False)

    if result.returncode != 0:
        raise GhostscriptError("Ghostscript failed to rasterize the document: ", result)

    return result


def split_png_images(data: bytes) -> List[bytes]:
    # PNG signature
    png_signature = b'\x89PNG\r\n\x1a\n'
    images = []

    # Find the first PNG signature
    start = data.find(png_signature)
    if start == -1:
        # No PNG images found in the data
        return []

    while True:
        # Find the next PNG signature in the data
        next_start = data.find(png_signature, start + 1)
        if next_start == -1:
            # No more images
            images.append(data[start:])
            break
        images.append(data[start:next_start])
        start = next_start

    return images


def rasterize_pdf_to_bytes(
    fp: Union[PathLike, str], *, dpi: int = 100, device="pnggray", downscale_factor: Optional[int] = None
) -> Dict[int, bytes]:
    if "png" not in device:
        raise ValueError("Device must be a PNG device for rasterize_pdf_to_bytes")

    result = rasterize_pdf(fp, "%stdout", dpi=dpi, device=device, downscale_factor=downscale_factor)

    images = split_png_images(result.stdout)

    return {idx: image for idx, image in enumerate(images, start=1)}


def rasterize_page_to_bytes(fp: Union[PathLike, str], idx: int, *, dpi: int = 200, device="pnggray") -> bytes:
    if isinstance(fp, bytes):
        with tempfile.NamedTemporaryFile(suffix=".pdf") as f:
            f.write(fp)
            f.flush()
            result = rasterize_page(f.file.name, "%stdout", idx, dpi=dpi, device=device)
    else:
        result = rasterize_page(fp, "%stdout", idx, dpi=dpi, device=device)

    return result.stdout


def rasterize_page_to_path(
    fp: Union[PathLike, str],
    idx: int,
    output_path: PathLike,
    *,
    dpi: int = 200,
    device="pnggray",
) -> Path:
    if isinstance(fp, bytes):
        with tempfile.NamedTemporaryFile(suffix=".pdf") as f:
            f.write(fp)
            f.flush()
            rasterize_page(f.file.name, str(output_path), idx, dpi=dpi, device=device)
    else:
        rasterize_page(fp, str(output_path), idx, dpi=dpi, device=device)

    return Path(output_path)


def compress_pdf(
    fp: Union[PathLike, str],  # Ghostscript insists on a file instead of bytes
    output_path: str,
    *,
    compression: Literal["jpeg", "lossless"] = "jpeg",
):
    compression_args = []
    if compression == "jpeg":
        compression_args = [
            "-dAutoFilterColorImages=false",
            "-dColorImageFilter=/DCTEncode",
            "-dAutoFilterGrayImages=false",
            "-dGrayImageFilter=/DCTEncode",
        ]
    elif compression == "lossless":
        compression_args = [
            "-dAutoFilterColorImages=false",
            "-dColorImageFilter=/FlateEncode",
            "-dAutoFilterGrayImages=false",
            "-dGrayImageFilter=/FlateEncode",
        ]
    else:
        compression_args = [
            "-dAutoFilterColorImages=true",
            "-dAutoFilterGrayImages=true",
        ]

    args_gs = (
        [
            GS,
            "-q",
            "-dBATCH",
            "-dNOPAUSE",
            "-dSAFER",
            "-dCompatibilityLevel=1.5",
            "-sDEVICE=pdfwrite",
            "-dAutoRotatePages=/None",
            "-sColorConversionStrategy=LeaveColorUnchanged",
        ]
        + compression_args
        + [
            "-dJPEGQ=95",
            "-dPDFA=2",
            "-dPDFACompatibilityPolicy=1",
            "-sOutputFile=" + output_path,
            str(fp),
        ]
    )

    result = run(args_gs, stdout=PIPE, stderr=PIPE, check=False)

    if result.returncode != 0:
        raise GhostscriptError("Ghostscript failed to compress the document", result)

    return result


def compress_pdf_to_bytes(fp: Union[PathLike, str], *, compression: Literal["jpeg", "lossless"] = "jpeg") -> bytes:
    result = compress_pdf(fp, output_path="%stdout", compression=compression)

    return result.stdout


def compress_pdf_to_path(
    fp: Union[PathLike, str],
    output_path: PathLike,
    *,
    compression: Literal["jpeg", "lossless"] = "jpeg",
) -> Path:
    compress_pdf(fp, output_path=str(output_path), compression=compression)

    return Path(output_path)



"""Top-level package for Docprompt."""

__author__ = """Frankie Colson"""
__email__ = 'frank@pageleaf.io'
__version__ = '0.1.0'


import docprompt.silenceable_tqdm  # noqa
from docprompt.schema.document import Document
from docprompt.schema.layout import NormBBox, TextBlock
from docprompt.schema.pipeline import DocumentCollection, DocumentNode, PageNode
from docprompt.utils import load_document

Document.model_rebuild()



from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, ContextManager, Optional

from docprompt.schema.operations import ProviderResult
from docprompt.service_providers.types import OPERATIONS

if TYPE_CHECKING:
    from docprompt.schema.document import Document


class BaseProvider(metaclass=ABCMeta):
    name: str

    @abstractmethod
    def _call(self, document: "Document", pages=list[int]) -> ProviderResult:
        raise NotImplementedError

    @property
    @abstractmethod
    def capabilities(self) -> list[OPERATIONS]:
        raise NotImplementedError

    def process_document(self, document: "Document", pages: Optional[list[int]] = None) -> ProviderResult:
        """
        Should return a ProviderResult object
        """
        pages = pages or list(range(1, document.num_pages + 1))
        return self._call(document, pages)



import re
from datetime import date
from typing import Optional

from dateutil.parser import ParserError
from dateutil.parser import parse as dateutil_parse

start_pattern = r"(?:^|\s)"
end_pattern = r"(?:[;,.]?\s|$)"


NUMERICAL_PATTERNS = [
    r"(?:^|\s)(\d{1,2}/\d{1,2}/\d{4})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{1,2}/\d{1,2}/\d{2})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{4}/\d{1,2}/\d{1,2})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{1,2}-\d{1,2}-\d{4})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{1,2}-\d{1,2}-\d{2})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{4}-\d{1,2}-\d{1,2})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{4}\.\d{1,2}\.\d{1,2})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{1,2}\.\d{1,2}\.\d{4})(?:[;,.]?\s|$)",
    r"(?:^|\s)(\d{1,2}\.\d{1,2}\.\d{2})(?:[;,.]?\s|$)",
]

TEXT_PATTERNS = [
    # Month name (full and abbreviated) followed by day and year
    r"(?:^|\s)((?:Jan(?:uary)?|Feb(?:ruary)?|Mar(?:ch)?|Apr(?:il)?|May|Jun(?:e)?|Jul(?:y)?|Aug(?:ust)?|Sep(?:tember)?|Oct(?:ober)?|Nov(?:ember)?|Dec(?:ember)?)\s+\d{1,2}(?:st|nd|rd|th)?,?\s+\d{4})(?:[;,.]?\s|$)",  # Added ordinal suffixes
    # Day followed by month name (full and abbreviated) and year
    r"(?:^|\s)(\d{1,2}(?:st|nd|rd|th)?\s+(?:Jan(?:uary)?|Feb(?:ruary)?|Mar(?:ch)?|Apr(?:il)?|May|Jun(?:e)?|Jul(?:y)?|Aug(?:ust)?|Sep(?:tember)?|Oct(?:ober)?|Nov(?:ember)?|Dec(?:ember)?)\s+\d{4})(?:[;,.]?\s|$)",  # Added ordinal suffixes
    # Year followed by month name (full and abbreviated) and day
    r"(?:^|\s)(\d{4}\s+(?:Jan(?:uary)?|Feb(?:ruary)?|Mar(?:ch)?|Apr(?:il)?|May|Jun(?:e)?|Jul(?:y)?|Aug(?:ust)?|Sep(?:tember)?|Oct(?:ober)?|Nov(?:ember)?|Dec(?:ember)?)\s+\d{1,2})(?:[;,.]?\s|$)",
]


DEFAULT_PATTERNS = NUMERICAL_PATTERNS + TEXT_PATTERNS


def register_date_pattern(pattern: str):
    """
    Registers a new date pattern to be used by the date extraction
    functions
    """
    if pattern not in DEFAULT_PATTERNS:
        DEFAULT_PATTERNS.append(pattern)


def get_date_pattern_regexes():
    return [re.compile(x, re.IGNORECASE) for x in DEFAULT_PATTERNS]


def get_date_matches_for_text(
    text: str, unique: bool = False, patterns: Optional[list[re.Pattern]] = None
) -> list[re.Match]:
    results = []

    patterns = patterns or get_date_pattern_regexes()

    for regex in patterns:
        for match in regex.finditer(text):
            results.append(match)

    if unique:
        return list(set(results))

    return results


def get_date_strings_for_text(
    text: str, unique: bool = False, patterns: Optional[list[re.Pattern]] = None
) -> list[str]:
    results = []

    patterns = patterns or get_date_pattern_regexes()

    for regex in patterns:
        for match_text in regex.findall(text):
            results.append(match_text.strip())

    if unique:
        return list(set(results))

    return results


def get_dates_for_text(text: str, unique: bool = False, patterns: Optional[list[re.Pattern]] = None) -> list[date]:
    results = []

    patterns = patterns or get_date_pattern_regexes()

    date_strings = get_date_strings_for_text(text, unique=unique, patterns=patterns)

    for date_string in date_strings:
        try:
            parsed = dateutil_parse(date_string)
            results.append(parsed.date())
        except ParserError:
            continue

    if unique:
        return list(set(results))

    return results



import hashlib
import tempfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import Optional, Union
from urllib.parse import unquote

import fsspec
import magic
from pikepdf import Pdf

from docprompt._exec.ghostscript import compress_pdf_to_path
from docprompt.schema.document import Document


def ensure_path(fp: Union[Path, PathLike]) -> Path:
    """
    Ensures that a file path is a Path object
    """
    if not isinstance(fp, Path):
        fp = Path(fp)

    return fp


def is_pdf(fd: Union[Path, PathLike, bytes]) -> bool:
    """
    Determines if a file is a PDF
    """
    if not isinstance(fd, bytes):
        with open(fd, "rb") as f:
            fd: bytes = f.read(1024)
            # We only need the first 1024 bytes to determine if it's a PDF

    mime = magic.from_buffer(fd, mime=True)

    return mime == "application/pdf"


def get_page_count(fd: Union[Path, PathLike, bytes]) -> int:
    """
    Determines the number of pages in a PDF
    """
    if not isinstance(fd, bytes):
        with open(fd, "rb") as f:
            fd = f.read()

    with Pdf.open(BytesIO(fd)) as pdf:
        return len(pdf.pages)


def load_document(
    fp: Union[Path, PathLike, bytes],
    *,
    file_name: Optional[str] = None,
    do_compress: bool = False,
    do_clean: bool = False,
) -> Document:
    """
    Loads a document from a file path
    """
    if isinstance(fp, bytes):
        file_bytes = fp
        if file_name is None:
            file_name = "document.pdf"
    else:
        if not isinstance(fp, Path):
            fp = Path(fp)

        if fp.is_symlink():
            fp = fp.resolve()

        file_name = fp.name

        with open(fp, "rb") as f:
            file_bytes: bytes = f.read()

    if not is_pdf(file_bytes):
        raise ValueError("File is not a PDF")

    if do_compress or do_clean:
        with tempfile.TemporaryDirectory(f"_process_{file_name}") as temp_dir:
            temp_path = Path(temp_dir)
            temp_file = temp_path / file_name

            with temp_file.open("wb") as f:
                f.write(file_bytes)

            if do_compress:
                compress_pdf_to_path(temp_file, temp_path / "compressed.pdf")
                file_bytes = (temp_path / "compressed.pdf").read_bytes()

            if do_clean:
                raise NotImplementedError("Cleaning with unpaper is not yet implemented")
                # compress_pdf_to_path(temp_file, temp_path / "cleaned.pdf", clean=True)
                # file_bytes = (temp_path / "cleaned.pdf").read_bytes()

    return Document(name=unquote(file_name), file_path=str(fp), file_bytes=file_bytes)


def load_document_from_url(url: str, **kwargs):
    with fsspec.open(url, "rb") as f:
        file_bytes: bytes = f.read()

    file_name = unquote(url.split("/")[-1])

    if not is_pdf(file_bytes):
        raise ValueError("File is not a PDF")

    return Document(name=file_name, file_path=url, file_bytes=file_bytes)


def load_documents_from_urls(urls: list[str], max_workers: int = 5, **kwargs) -> list[Document]:
    documents = []

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_url = {executor.submit(load_document_from_url, url): url for url in urls}
        for future in as_completed(future_to_url):
            url = future_to_url[future]
            try:
                documents.append(future.result())
            except Exception as exc:
                print(f"{url} generated an exception: {exc}")
                raise

    return documents


def hash_from_bytes(byte_data: bytes) -> str:
    stream = BytesIO(byte_data)

    hash = hashlib.md5()
    b = bytearray(128 * 1024)
    mv = memoryview(b)

    while n := stream.readinto(mv):
        hash.update(mv[:n])

    return hash.hexdigest()



import os

import tqdm


class SilenceableTqdm(tqdm.tqdm):
    """
    Wrapper for tqdm that disables all progress bars if DOCPROMPT_PROGRESS_BARS is set to a falsey value
    ("0", "False", "FALSE", "false").

    Note: this check is done every time a tqdm iterator is initialized, so normally for each method run. Therefore
    progress bars can be enabled and disabled at runtime, but not during a specific iteration.
    """

    def __init__(self, *args, **kwargs):
        """
        Passes `disable=True` to tqdm if `self.no_progress_bars` is set to True.
        """
        if self.no_progress_bars:
            kwargs["disable"] = True

        super().__init__(*args, **kwargs)

    @property
    def no_progress_bars(self):
        """
        Reads the DOCPROMPT_PROGRESS_BARS env var to check if the progress bars should be disabled.
        """
        return os.getenv("DOCPROMPT_PROGRESS_BARS", "1") in ["0", "False", "FALSE", "false"]

    @property
    def disable(self):
        return self.no_progress_bars or self._disable

    @disable.setter
    def disable(self, value):
        self._disable = value


tqdm.std.tqdm = SilenceableTqdm
tqdm.tqdm = SilenceableTqdm



from typing import Optional, Union

from pydantic import BaseModel, Field, PositiveInt, computed_field

from .layout import TextBlock


class PageTextExtractionOutput(BaseModel):
    """
    Represents the output of text extraction for a single page
    """

    text: str = Field(description="The provider-sourced text for the page")

    words: list[TextBlock] = Field(default_factory=list, description="The provider-sourced words for the page")
    lines: list[TextBlock] = Field(default_factory=list, description="The provider-sourced lines for the page")
    blocks: list[TextBlock] = Field(default_factory=list, description="The provider-sourced blocks for the page")

    @computed_field
    @property
    def word_count(self) -> PositiveInt:
        return len(self.words)

    @computed_field
    @property
    def average_word_confidence(self) -> Union[float, None]:
        confidences = [word.confidence for word in self.words if word.confidence is not None]

        if len(confidences) == 0:
            return None

        return sum(confidences) / len(confidences)


class PageResult(BaseModel):
    """
    Represents the result of processing a single page
    (page-wise)
    """

    provider_name: str
    page_number: int

    ocr_result: Optional[PageTextExtractionOutput] = None


class DocumentResult(BaseModel):
    """
    Represents the result of operations that consider the
    document as a whole (document-wise)

    Examples:

    * Multi-page DocVQA
    """

    provider_name: str


class ProviderResult(BaseModel):
    """
    Represents the result of processing a document
    """

    provider_name: str

    page_results: Optional[list[PageResult]] = None
    document_result: Optional[DocumentResult] = None



from enum import Enum
from typing import Dict

from pydantic import BaseModel, Field

from docprompt.schema.layout import TextBlock


class OPERATIONS(Enum):
    IMAGE_PROCESSING = "image_processing"
    LAYOUT_ANALYSIS = "layout_analysis"
    TEXT_EXTRACTION = "text_extraction"



from typing import Generic, TypeVar

from pydantic import BaseModel, Field

from .document import Document

DocumentCollectionMetadata = TypeVar("DocumentCollectionMetadata", bound=BaseModel)
DocumentNodeMetadata = TypeVar("DocumentNodeMetadata", bound=BaseModel)
PageNodeMetadata = TypeVar("PageNodeMetadata", bound=BaseModel)


class PageNode(BaseModel, Generic[PageNodeMetadata]):
    """
    Represents a single page in a document, with some metadata
    """

    document: "DocumentNode" = Field(exclude=True)
    metadata: PageNodeMetadata


class DocumentNode(BaseModel, Generic[DocumentNodeMetadata, PageNodeMetadata]):
    """
    Represents a single document, with some metadata
    """

    document: Document
    page_nodes: list[PageNode[PageNodeMetadata]]
    metadata: DocumentNodeMetadata

    @property
    def file_hash(self):
        return self.document.document_hash

    @property
    def document_name(self):
        return self.document.name


class DocumentCollection(BaseModel, Generic[DocumentCollectionMetadata, DocumentNodeMetadata, PageNodeMetadata]):
    """
    Represents a collection of documents with some common metadata
    """

    document_nodes: list[DocumentNode[DocumentNodeMetadata, PageNodeMetadata]]
    metadata: DocumentCollectionMetadata
